(module matlab)

;;; TODO Scheme to matlab conversion and proper matlab function calls
;;; TODO Matlab cell, struct, logical and function 
;;; TODO Support for building matlab structures
;;; TODO Support for complex numbers

(include "QobiScheme.sch")
(include "c-macros.sch")

(c-include "matlab.h")
(c-include "engine.h")

(define *matlab-class-unknown*  (c-value int "mxUNKNOWN_CLASS"))
(define *matlab-class-cell*     (c-value int "mxCELL_CLASS"))
(define *matlab-class-struct*   (c-value int "mxSTRUCT_CLASS"))
(define *matlab-class-logical*  (c-value int "mxLOGICAL_CLASS"))
(define *matlab-class-char*     (c-value int "mxCHAR_CLASS"))
(define *matlab-class-double*   (c-value int "mxDOUBLE_CLASS"))
(define *matlab-class-single*   (c-value int "mxSINGLE_CLASS"))
(define *matlab-class-int8*     (c-value int "mxINT8_CLASS"))
(define *matlab-class-uint8*    (c-value int "mxUINT8_CLASS"))
(define *matlab-class-int16*    (c-value int "mxINT16_CLASS"))
(define *matlab-class-uint16*   (c-value int "mxUINT16_CLASS"))
(define *matlab-class-int32*    (c-value int "mxINT32_CLASS"))
(define *matlab-class-uint32*   (c-value int "mxUINT32_CLASS"))
(define *matlab-class-int64*    (c-value int "mxINT64_CLASS"))
(define *matlab-class-uint64*   (c-value int "mxUINT64_CLASS"))
(define *matlab-class-function* (c-value int "mxFUNCTION_CLASS"))

(define *matlab-sizeof-mwSize* (c-sizeof "mwSize"))

;;; Matlab engine bindings

(define matlab-start (c-function pointer ("engOpen" string)))
(define matlab-stop (c-function int ("engClose" pointer)))
(define matlab-variable! 
  (c-function pointer ("engGetVariable" pointer string)))
(define matlab-variable-set!
  (c-function int ("engPutVariable" pointer string pointer)))
(define matlab-eval-string (c-function int ("engEvalString" pointer string)))
(define matlab-set-output-buffer 
  (c-function int ("engOutputBuffer" pointer pointer int)))
(define matlab-set-visible 
  (c-function int ("engSetVisible" pointer int)))
(define matlab-get-visible 
  (c-function int ("engGetVisible" pointer pointer)))


(define *default-matlab-engine-command* "matlab -nosplash -nodesktop -nojvm")
(define *default-matlab-buffer-size* 400)

(define (with-matlab-default-engine f) 
  (with-matlab-engine *default-matlab-engine-command* f))

(define (with-matlab-engine str f)
  (let* ((engine (matlab-start str))
	 (result (f engine)))
    (matlab-stop engine)
    result))

(define (with-matlab-default-output-buffer engine f)
  (with-matlab-output-buffer engine *default-matlab-buffer-size* f))

(define (with-matlab-output-buffer engine size f)
  (with-alloc 
   size 
   (lambda (buffer)
     (matlab-set-output-buffer engine buffer size)
     (let ((result (f (lambda () (c-string->string buffer)))))
       (matlab-set-output-buffer engine null 0)
       result))))


;;; Matlab MX Array bindings

(define matlab-class-id (c-function int ("mxGetClassID" pointer)))
(define matlab-class-name (c-function string ("mxGetClassName" pointer)))
(define matlab-array->string (c-function string ("mxArrayToString" pointer)))
(define matlab-calloc (c-function pointer ("mxCalloc" int int)))
(define matlab-malloc (c-function pointer ("mxMalloc" int)))
(define matlab-free (c-function void ("mxFree" pointer)))
(define matlab-make-numeric-matrix 
  (c-function pointer ("mxCreateNumericMatrix" int int int int)))
(define matlab-destroy (c-function void ("mxDestroyArray" pointer)))
(define matlab-nr-rows (c-function int ("mxGetM" pointer)))
(define matlab-nr-columns (c-function int ("mxGetN" pointer)))
(define matlab-nr-dimensions 
  (c-function int ("mxGetNumberOfDimensions" pointer)))
(define matlab-nr-elements (c-function int ("mxGetNumberOfElements" pointer)))
(define matlab-dimensions-internal 
  (c-function pointer ("mxGetDimensions" pointer)))
(define matlab-data-set! (c-function void ("mxSetData" pointer pointer)))
(define matlab-data (c-function void ("mxGetData" pointer)))
(define matlab-real-double-data (c-function pointer ("mxGetPr" pointer)))
(define matlab-real-double-data-set! 
  (c-function void ("mxSetPr" pointer pointer)))
(define matlab-complex-double-data (c-function pointer ("mxGetPi" pointer)))
(define matlab-complex-double-data-set! 
  (c-function void ("mxSetPi" pointer pointer)))
(define matlab-element-size (c-function int ("mxGetElementSize" pointer)))
(define matlab-calculate-index-internal
  (c-function int ("mxCalcSingleSubscript" pointer int pointer)))

(define (matlab-dimensions matrix)
  (c-exact-array->list (matlab-dimensions-internal matrix)
		       *matlab-sizeof-mwSize*
		       (matlab-nr-dimensions matrix)
		       #f))

(define (matlab-calculate-index matrix array subscripts)
  (matlab-calculate-index-internal
   matrix
   (length subscripts)
   (list->c-array array subscripts *matlab-sizeof-mwSize* #f)))

(define (matlab-string? m)
  (= *matlab-class-char* (matlab-class-id m)))

(define (matlab-matrix-exact? m)
  (eq? (car (matlab-matrix-numeric-type m)) 'exact))
(define (matlab-matrix-inexact? m)
  (eq? (car (matlab-matrix-numeric-type m)) 'inexact))
(define (matlab-matrix-signed? m)
  (eq? (cadr (matlab-matrix-numeric-type m)) 'signed))
(define (matlab-matrix-unsigned? m)
  (eq? (cadr (matlab-matrix-numeric-type m)) 'unsigned))

(define (matlab-matrix-numeric-type m)
  (let ((id (matlab-class-id m)))
    (cond
     ((or (= *matlab-class-double* id)
	  (= *matlab-class-single* id))
      `(inexact signed))
     ((or (= *matlab-class-int8* id)
	  (= *matlab-class-int16* id)
	  (= *matlab-class-int32* id)
	  (= *matlab-class-int64* id))
      `(exact signed))
     ((or (= *matlab-class-uint8* id)
	  (= *matlab-class-uint16* id)
	  (= *matlab-class-uint32* id)
	  (= *matlab-class-uint64* id))
      `(exact unsigned))
     (else '(#f #f)))))

;;; Various handy utilities

(define (list->c-array array l set-element element-size)
  (map-indexed (lambda (x i) 
		 (set-element
		  array
		  (* i element-size)
		  x)) l)
  array)

(define (list->c-inexact-array array l element-size signed?)
  (list->c-array
   array
   l
   (c-sized-inexact-ptr-set! element-size signed?) 
   element-size))

(define (list->c-exact-array array l element-size signed?)
  (list->c-array
   array
   l
   (c-sized-int-ptr-set! element-size signed?) 
   element-size))

(define (with-array elements element-size f)
  (with-alloc
   (* elements element-size)
   f))

(define (c-array->list array get-element element-size nr-elements)
  (vector->list (c-array->vector array get-element element-size nr-elements)))

(define (c-array->vector array get-element element-size nr-elements)
  (map-n-vector 
   (lambda (x) (get-element array (* x element-size))) nr-elements))

(define (c-exact-array->list array element-size nr-elements signed?)
  (vector->list 
   (c-exact-array->vector array element-size nr-elements signed?)))

(define (c-exact-array->vector array element-size nr-elements signed?)
  (c-array->vector
   array
   (c-sized-int-ptr-ref element-size signed?) element-size nr-elements))

(define (c-inexact-array->list array element-size nr-elements signed?)
  (vector->list 
   (c-inexact-array->vector array element-size nr-elements signed?)))

(define (c-inexact-array->vector array element-size nr-elements signed?)
  (c-array->vector
   array
   (c-sized-inexact-ptr-ref element-size signed?) element-size nr-elements))

(define (c-ptr-byte-offset ptr off)
  ((lap (ptr off)
	(POINTER_TSCP ("add" 
		       ("(char*)" (TSCP_POINTER ptr)) 
		       (TSCP_S2CINT off))))
   ptr off))

(define (c-sized-int-ptr-ref size signed?)
  (cond
   ((= size 1) c-byte-ref)
   ((= size c-sizeof-short) (if signed? c-shortint-ref c-shortunsigned-ref))
   ((= size c-sizeof-int)   (if signed? c-int-ref c-unsigned-ref))
   ((= size c-sizeof-long)  (if signed? c-longint-ref c-longunsigned-ref))
   (else (fuck-up))))

(define (c-sized-int-ptr-set! size signed?)
  (cond 
   ((= size 1) c-byte-set!)
   ((= size c-sizeof-short) (if signed? c-shortint-set! c-shortunsigned-set!))
   ((= size c-sizeof-int)   (if signed? c-int-set! c-unsigned-set!))
   ((= size c-sizeof-long)  (if signed? c-longint-set! c-longunsigned-set!))
   (else (fuck-up))))

(define (c-sized-inexact-ptr-ref size signed?)
  (cond
   ((= size c-sizeof-float) c-float-ref)
   ((= size c-sizeof-double) c-double-ref)
   (else (fuck-up))))

(define (c-sized-inexact-ptr-set! size signed?)
  (cond 
   ((= size c-sizeof-float) c-float-set!)
   ((= size c-sizeof-double) c-double-set!)
   (else (fuck-up))))

(define (matlab-size-ref struct x)
  ((c-sized-int-ptr-ref *matlab-sizeof-mwSize* #f)
   struct (* *matlab-sizeof-mwSize* x)))

(define (matlab-matrix->list matrix signed?)
  (vector->list (matlab-matrix->matrix matrix signed?)))

(define (matlab-matrix->matrix matrix signed?)
  (let ((size (matlab-element-size matrix)))
    (matlab-data->vector 
     (cond ((matlab-matrix-exact? matrix) (matlab-data matrix))
	   ((matlab-matrix-inexact? matrix) (matlab-real-double-data matrix))
	   (else (fuck-up)))
     (cond ((matlab-matrix-exact? matrix)
	    (lambda (array nr-elements) 
	      (c-exact-array->vector array size nr-elements signed?)))
	   ((matlab-matrix-inexact? matrix)
	    (lambda (array nr-elements) 
	      (c-inexact-array->vector array size nr-elements signed?)))
	   (else (fuck-up)))
     (matlab-element-size matrix) 
     (reverse (matlab-dimensions matrix)))))

(define (matlab-data->vector data get-row element-size dimensions)
  (cond
   ((= 1 (length dimensions))
    (get-row data (car dimensions)))
   ((< 1 (length dimensions))
    (transpose
     (map-n-vector
      (lambda (n)
	(matlab-data->vector 
	 (c-ptr-byte-offset data 
			    (* n element-size (apply * (cdr dimensions))))
	 get-row element-size (cdr dimensions)))
      (car dimensions))))
   (else (fuck-up))))


(define (scheme->matlab s)
  (fuck-up))
(define (matlab->scheme m)
  (cond ((or (matlab-matrix-exact? m) (matlab-matrix-inexact? m))
	 (matlab-matrix->matrix m (matlab-matrix-signed? m)))
	((or (matlab-string? m)) (matlab-array->string m))
	(else (fuck-up))))

(define (with-matlab-variable engine var f)
  (let* ((x (matlab-variable! engine var))
	 (result (f x)))
    (matlab-destroy x)
    result))

(define (matlab-show-variable engine name)
  (with-matlab-variable 
   engine 
   name
   (lambda (var)
     (display (format "~a is ~a~%" name (matlab->scheme var))))))

(define (test-matlab)
  (with-matlab-default-engine 
   (lambda (engine)
     (with-matlab-default-output-buffer 
      engine
      (lambda (matlab-result-string)
	(matlab-eval-string engine "A=[[1,2,3];[4,5,6]]")
	(matlab-eval-string engine "B='Hi'")
	(matlab-eval-string engine "C=1")
	(matlab-show-variable engine "A")
	(matlab-show-variable engine "B")
	(matlab-show-variable engine "C")
	(matlab-eval-string engine "D=randn(2,2,2,2)")
	(display (matlab-result-string))
	(matlab-show-variable engine "D")
	(matlab-eval-string engine "'Can also get the output from matlab'")
	(display (matlab-result-string))
	(newline)
	(display "Bye")(newline))))))

(define malloc (c-function pointer ("malloc" int)))
(define memcpy (c-function pointer ("memcpy" pointer pointer int)))
(define bzero (c-function void ("bzero" pointer int)))
(define free (c-function void ("free" pointer)))
(define (with-alloc x f)
  (let* ((data (malloc x))
	 (r (begin (bzero data x) (f data))))
    (free data)
    r))
