#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <signal.h>
#include <termios.h>
#include <sys/poll.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/wait.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/rfcomm.h>

/*
  This code is in part lifted from rfcomm in bluez-utils.3.36
 */

static int estr2ba(char *str, bdaddr_t *ba)
{
	/* Only trap "local", "any" is already dealt with */
	if(!strcmp(str, "local")) {
		bacpy(ba, BDADDR_LOCAL);
		return 0;
	}
	return str2ba(str, ba);
}

int* bluetooth_connect(int dev_id, char* bluetooth_addr)
{
  struct sockaddr_rc laddr;
  laddr.rc_family = AF_BLUETOOTH;
  bacpy(&laddr.rc_bdaddr, BDADDR_ANY);
  laddr.rc_channel = 0;

  struct sockaddr_rc raddr;
  raddr.rc_family = AF_BLUETOOTH;
  str2ba(bluetooth_addr, &raddr.rc_bdaddr);
  raddr.rc_channel = 1;

  int control_socket = socket(AF_BLUETOOTH, SOCK_RAW, BTPROTO_RFCOMM);
  if (control_socket < 0) {
    error("Can't open rfcomm control socket : %s\n", strerror(errno));
    exit(1);
  }
  
  int bluetooth_socket = 0;

  while(raddr.rc_channel < 10)
  {
    bluetooth_socket = socket(AF_BLUETOOTH, SOCK_STREAM, BTPROTO_RFCOMM);
    if(bluetooth_socket < 0 ||
       bind(bluetooth_socket, (struct sockaddr *) &laddr, sizeof(laddr)) < 0)
    {
      printf("Bluetooh socket is busy : %s\n", strerror(errno));
      exit(1);
    } 

    struct linger l = { .l_onoff = 1, .l_linger = 0 };
    
    if (setsockopt(bluetooth_socket, SOL_SOCKET, SO_LINGER, &l, sizeof(l)) < 0) {
      printf("Setting linger options on bluetooth socket failed : %s\n", 
	     strerror(errno));
      exit(1);
    }

    int status;
    if(status = connect(bluetooth_socket, (struct sockaddr *) &raddr, sizeof(raddr)) < 0)
    {
      if(errno == EBUSY)
      {
	printf("rfcomm socket is busy\n");
	close(bluetooth_socket);
	bluetooth_socket = 0;
	raddr.rc_channel++;
	continue;
      } else {
	printf("Can't connect to the rfcomm socket : %s\n", strerror(errno));
	close(bluetooth_socket);
	exit(1);
      }
    }
    else
      break;
  }

  if(!bluetooth_socket)
  {
    printf("rfcomm is busy\n"); 
    close(bluetooth_socket);
    exit(1);
  }

  socklen_t alen;
  alen = sizeof(laddr);
  if (getsockname(bluetooth_socket, (struct sockaddr *)&laddr, &alen) < 0) {
    printf("Can't get rfcomm socket name : %s\n", strerror(errno));
    close(bluetooth_socket);
    exit(1);
  }
  
  struct rfcomm_dev_req req;
  memset(&req, 0, sizeof(req));
  req.dev_id = dev_id;
  req.flags = (1 << RFCOMM_REUSE_DLC) | (1 << RFCOMM_RELEASE_ONHUP);

  bacpy(&req.src, &laddr.rc_bdaddr);
  bacpy(&req.dst, &raddr.rc_bdaddr);
  req.channel = raddr.rc_channel;
  
  
  int dev = ioctl(bluetooth_socket, RFCOMMCREATEDEV, &req);
  if (dev < 0) {
    printf("Can't create RFCOMM TTY");
    close(bluetooth_socket);
    exit(1);
  }

  char devname[MAXPATHLEN];
  snprintf(devname, MAXPATHLEN - 1, "/dev/rfcomm%d", dev);
  int fd, try = 30;

  while ((fd = open(devname, O_RDONLY | O_NOCTTY)) < 0) {
    if (errno == EACCES) {
      perror("Can't open RFCOMM device");
      goto release;
    }
    
    snprintf(devname, MAXPATHLEN - 1, "/dev/bluetooth/rfcomm/%d", dev);
    if ((fd = open(devname, O_RDONLY | O_NOCTTY)) < 0) {
      if (try--) {
	snprintf(devname, MAXPATHLEN - 1, "/dev/rfcomm%d", dev);
	usleep(100 * 1000);
	continue;
      }
      perror("Can't open RFCOMM device");
      goto release;
    }
  }

  close(bluetooth_socket);
  int* result = malloc(sizeof(int));
  *result = fd;
  return result;
  

release:
  memset(&req, 0, sizeof(req));
  req.dev_id = dev;
  req.flags = (1 << RFCOMM_HANGUP_NOW);
  ioctl(control_socket, RFCOMMRELEASEDEV, &req);
  
  close(bluetooth_socket);
  close(control_socket);
  return 0;
}

int get_bluetooth_fd(int* fd)
{
  return *fd;
}

void bluetooth_disconnect(int* fd)
{
  close(*fd);
}
